var RnaNormalize = function()
{
	var normalize = {};

	// The minial volume of RNA we want to take when normalizing
	// For RNA with high concentration we must make a larger mix to reach the target concentation
	var MINIMAL_RNA_VOLUME = 1; // µl
	var hasInitProtocols = false;
	
	var selectedRna;
	var subtypeRna = null;
	
	var autoGoNext = 0;
	var newRna = [];
	
	// Page initialization
	normalize.initPage = function()
	{
		// Step 1
		Events.addEventHandler('step-1', 'wizard-validate', normalize.validateStep1);
		Buttons.addClickHandler('btnSelectRna', normalize.selectRna);
		Events.addEventHandler('rna', 'base-selected', normalize.onRnaSelected);
		
		// Step 2
		Events.addEventHandler('normalizationProtocol', 'change', normalize.protocolOnChange);
		Events.addEventHandler('normalizationProtocol', 'keyup', normalize.protocolOnChange);
		Events.addEventHandler('step-2', 'wizard-initialize', normalize.initializeStep2);
		Events.addEventHandler('step-2', 'wizard-validate', normalize.validateStep2);

		// Step 3
		Events.addEventHandler('step-3', 'wizard-initialize', normalize.initializeStep3);
		Events.addEventHandler('normalizationDate', 'blur', Wizard.validateDate);
		
		// Navigation
		Buttons.addClickHandler('gocancel', Wizard.cancelWizard);
		Buttons.addClickHandler('gorestart', Wizard.restartWizard);
		Buttons.addClickHandler('gonext', Wizard.goNextOnClick);
		Buttons.addClickHandler('goprint', normalize.printVersion);
		Buttons.addClickHandler('goregister', Wizard.goRegister);

		// Final registration
		Events.addEventHandler('wizard', 'wizard-submit', normalize.submit);

		var rna = Data.json('page-data', 'rna-id');
		if (rna.length > 0)
		{
			Doc.show('step-1');
			autoGoNext = rna.length;
			// Get more information about the selected rna
			var url = '../Extraction.servlet?ID='+App.getSessionId();
			url += '&cmd=GetRNAInfoForNormalize&rnaIds=' + rna.join(',');
			Wizard.asyncJsonRequest(url, normalize.onRnaInfoLoaded);
		}
		else
		{
			var url = '../Extraction.servlet?ID='+App.getSessionId();
			url += '&cmd=GetRNAToNormalize';
			Wizard.showLoadingAnimation('Loading RNA extracts...');
			Wizard.asyncJsonRequest(url, normalize.onRnaLoaded);
		}
	}
	
	normalize.onRnaLoaded = function(response)
	{
		var frm = document.forms['reggie'];
		var rnaExtracts = response.rnaExtracts;
		
		rnaExtracts.sort(normalize.sortByQiacubePosition);
		
		if (rnaExtracts.length > 0)
		{
			var qiacubeRun = rnaExtracts[0].QiacubeDate + '#' + rnaExtracts[0].QiacubeRunNo;
			for (var i=0; i < rnaExtracts.length; i++)
			{
				var rna = rnaExtracts[i];
				var selected = (rna.QiacubeDate + '#' + rna.QiacubeRunNo) == qiacubeRun;
				normalize.addRnaOption(frm.rna, rna, selected);
			}
		}
		else
		{	
			Wizard.setInputStatus('rna', 'invalid', 'No RNA available for normalization.');
		}
		
		Doc.show('step-1');
		Doc.show('gonext');
	}
	
	/**
		Sort by qiacube date, run number and position
	*/
	normalize.sortByQiacubePosition = function(a, b)
	{
		if (a.QiacubeDate == b.QiacubeDate)
		{
			if (a.QiacubeRunNo == b.QiacubeRunNo)
			{
				return a.QiacubePosition-b.QiacubePosition;
			}
			else
			{
				return a.QiacubeRunNo - b.QiacubeRunNo;
			}
		}
		else
		{
			return a.QiacubeDate < b.QiacubeDate ? -1 : 1;
		}

		return a.id - b.id;
	}


	/**
		Sort by qiacube date, run number and position
	*/
	normalize.sortByStorageLocation = function(a, b)
	{
		if (a.bioWell.bioPlate.name == b.bioWell.bioPlate.name)
		{
			return a.bioWell.location < b.bioWell.location ? -1 : 1;
		}
		else
		{
			return a.bioWell.bioPlate.name < b.bioWell.bioPlate.name ? -1 : 1;
		}
	
		return a.id - b.id;
	}

	
	normalize.selectRna = function()
	{
		var frm = document.forms['reggie'];
		if (frm.rna.disabled) return;
		
		if (subtypeRna == null)
		{
			subtypeRna = Reggie.getSubtypeInfo('RNA');
		}
		
		newRna = [];
		var url = '&resetTemporary=1';
		url += '&tmpfilter:INT:itemSubtype='+subtypeRna.id;
		Dialogs.selectItem('EXTRACT', 'rna', 1, url);
	}

	normalize.onRnaSelected = function(event)
	{
		var rna = event.detail;
		
		var frm = document.forms['reggie'];
		var rnaList = frm.rna;
		var isNew = true;
		for (var i = 0; i < rnaList.length; i++)
		{
			if (rnaList[i].value == rna.id)
			{
				rnaList[i].selected = true;
				isNew = false;
			}
		}
		
		if (isNew) newRna[newRna.length] = rna.id;
		
		if (event.detail.remaining == 0 && newRna.length > 0)
		{
			// Get more information about the selected rna
			var url = '../Extraction.servlet?ID='+App.getSessionId();
			url += '&cmd=GetRNAInfoForNormalize&rnaIds=' + newRna.join(',');
			Wizard.asyncJsonRequest(url, normalize.onRnaInfoLoaded);
		}
	}

	normalize.onRnaInfoLoaded = function(response)
	{
		var frm = document.forms['reggie'];
		var rnaExtracts = response.rnaExtracts;
		rnaExtracts.sort(normalize.sortByQiacubePosition);
		
		for (var rnaNo = 0; rnaNo < rnaExtracts.length; rnaNo++)
		{
			var rna = rnaExtracts[rnaNo];
			normalize.addRnaOption(frm.rna, rna, true);
		}
		
		if (frm.rna.length == autoGoNext)
		{
			Wizard.goNext(true);
		}
	}
	
	normalize.addRnaOption = function(list, rna, selected)
	{
		var name = rna.name;
		if (rna.bioWell)
		{
			name += ' -- ' + Reggie.reformatDate(rna.QiacubeDate) + '; Run ' + rna.QiacubeRunNo + '; Pos ' + rna.QiacubePosition;
		}

		var option = new Option(name, rna.id, selected, selected);
		if (rna.specimen && rna.specimen.YellowLabel != null) option.className = 'yellow';
		option.rna = rna;
		list[list.length] = option;
	}
	
	normalize.validateStep1 = function(event)
	{
		var frm = document.forms['reggie'];
		var rnaList = frm.rna;
		
		selectedRna = [];
		for (var i=0; i < rnaList.length; i++)
		{
			if (rnaList[i].selected)
			{
				selectedRna[selectedRna.length] = rnaList[i].rna;
			}
		}
			
		if (selectedRna.length == 0)
		{
			Wizard.setInputStatus('rna', 'invalid', 'Not selected', 'invalid');
			event.preventDefault();
		}
		else
		{
			Wizard.setInputStatus('rna', 'valid');
		}
	}


	normalize.initializeStep2 = function()
	{
		var frm = document.forms['reggie'];
		// Check if all RNA from same Qiacube run
		var qiacubeRun = selectedRna[0].QiacubeDate + '#' + selectedRna[0].QiacubeRunNo;
		var sameQiacubeRun = true;
		for (var i=1; i < selectedRna.length; i++)
		{
			var rna = selectedRna[i];
			if (qiacubeRun != rna.QiacubeDate+'#'+rna.QiacubeRunNo)
			{
				sameQiacubeRun = false;
				break;
			}
		}
		
		if (sameQiacubeRun)
		{
			selectedRna.sort(normalize.sortByQiacubePosition);
			Doc.show('qiacube-info');
			Doc.element('qiacube-date').innerHTML = Reggie.reformatDate(selectedRna[0].QiacubeDate);
			Doc.element('qiacube-run').innerHTML = selectedRna[0].QiacubeRunNo;
		}
		else
		{
			selectedRna.sort(normalize.sortByStorageLocation);
		}
		
		var html = '<table id="extractTable">';
		html += '<thead class="bg-filled-100">';
		// First header row
		html += '<tr>';
		html += '<th>RNA</th>';
		if (sameQiacubeRun)
		{
			html += '<th>Qiacube</th>';
		}
		html += '<th>Storage</th>';
		html += '<th>NDConc</th>';
		html += '<th>Remain</th>';	
		html += '<th colspan="2">Volumes</th>';
		html += '<th></th>';
		html += '</tr>';
		// Second header row
		html += '<tr>';
		html += '<th></th>';
		if (sameQiacubeRun)
		{
			html += '<th>position</th>';
		}
		html += '<th>box</th>';
		html += '<th>(ng/µl)</th>';
		html += '<th>(µg)</th>';
		html += '<th>(µl)</th>';
		html += '<th>H<sub>2</sub>O</th>';
		html += '<th>Remarks</th>';
		html += '</tr>';
		html += '</thead>';
		html += '<tbody>';
		
		var expectedQiacubePosition = 0;
		var homeUrl = Data.get('page-data', 'home-url');
		var yellowImg = '<img src="'+homeUrl+'/images/yellow-label.png">';
		for (var i=0; i < selectedRna.length; i++)
		{
			var rna = selectedRna[i];
			var isYellow = rna.specimen && rna.specimen.YellowLabel != null;
			if (sameQiacubeRun)
			{
				expectedQiacubePosition++;
				while (expectedQiacubePosition != rna.QiacubePosition && expectedQiacubePosition <= 12)
				{
					html += normalize.insertEmpty(expectedQiacubePosition);
					expectedQiacubePosition++;
				}
			}

			var img = isYellow ? yellowImg : '';
			html += '<tr class="highlight'+(isYellow ? ' yellow-specimen' : '')+'">';
			html += '<td class="extract if-yellow">'+img+Strings.encodeTags(rna.nextAliquotName)+'</td>';
			if (sameQiacubeRun)
			{
				html += '<td class="qiacubepos">'+rna.QiacubePosition+'</td>';
			}
			html += '<td class="box">'+Strings.encodeTags(rna.bioWell.bioPlate.name + ' ' + rna.bioWell.location)+'</td>';
			html += '<td class="ndconc">'+ Numbers.formatNumber(rna.NDConc, 1) + '</td>';
			html += '<td class="remain">'+ Numbers.formatNumber(rna.remainingQuantity, 2) + '</td>';
			html += '<td class="volume" id="rnaVolume.'+rna.id+'"></td>';
			html += '<td class="water" id="rnaWater.'+rna.id+'"></td>';
			html += '<td class="remarks" id="remarks.'+rna.id+'"></td>';
			html += '</tr>'
		}
		if (sameQiacubeRun)
		{
			expectedQiacubePosition++;
			while (expectedQiacubePosition <= 12)
			{
				html += normalize.insertEmpty(expectedQiacubePosition);
				expectedQiacubePosition++;
			}
		}
		html += '</tbody>';
		html += '</table>';
		
		Doc.element('rnaNormalize').innerHTML = html;
		
		Wizard.setCurrentStep(2);
		Doc.show('gonext');
		Doc.show('goprint');
		Doc.show('gocancel');
		Wizard.keepSessionAlive();
		Reggie.loadProtocols('RNA_NORMALIZATION_PROTOCOL', 'normalizationProtocol', 'RNA_TARGET_AMOUNT,RNA_MINIMAL_AMOUNT,RNA_TARGET_VOLUME');
		frm.normalizationProtocol.focus();
	}
	
	normalize.protocolOnChange = function()
	{
		var frm = document.forms['reggie'];

		if (!hasInitProtocols)
		{
			hasInitProtocols = true;
			if (frm.normalizationProtocol.length == 0)
			{
				Wizard.setFatalError('No RNA normalization protocols found. Please ask the server administrator to create at least one protocol.');
				return;
			}
			for (var i=0; i < frm.normalizationProtocol.length; i++)
			{
				var p = frm.normalizationProtocol[i].item;
				
				var unit = 'µg';
				var factor = 1;
				if (p.RNATargetAmount <= 0.5)
				{
					unit = 'ng';
					factor = 1000;
				}
				
				var text = p.name + ' (';
				if (p.RNAMinimalAmount) text += Reggie.formatNumber(p.RNAMinimalAmount * factor, '', 1) + '-';
				text += p.RNATargetAmount ? Reggie.formatNumber(p.RNATargetAmount * factor, '', 1) : '? ';
				text += unit;
				text += ' in '+(p.RNATargetVolume || '? ')+'µl)';
				frm.normalizationProtocol[i].text = text;
			}
		}

		Wizard.showGoNextConfirmation(true, 'Print the lab protocol before continuing with the registration!');
		var protocol = frm.normalizationProtocol[frm.normalizationProtocol.selectedIndex].item;
		if (!protocol || !protocol.RNATargetAmount || !protocol.RNATargetVolume)
		{
			protocol = {};
			if (!protocol.RNATargetAmount) protocol.RNATargetAmount = 1.1;
			if (!protocol.RNATargetVolume) protocol.RNATargetVolume = 50;
			
			var msg = 'Missing normalization data; assuming ';
			msg += protocol.RNATargetAmount+ 'µg in '+protocol.RNATargetVolume+'µl.';
			Wizard.setInputStatus('normalizationProtocol', 'warning',  msg);
		}
		else
		{
			Wizard.setInputStatus('normalizationProtocol', 'valid');
		}
		if (!protocol.RNAMinimalAmount) protocol.RNAMinimalAmount = protocol.RNATargetAmount;
		Doc.element('normalization-protocol').innerHTML =  Strings.encodeTags(frm.normalizationProtocol[frm.normalizationProtocol.selectedIndex].text);

		// The concentration of the normalized RNA
		var normalizedConc = protocol.RNATargetAmount / protocol.RNATargetVolume; // µg/µl
		
		for (var i=0; i < selectedRna.length; i++)
		{
			var rna = selectedRna[i];
			
			// Aim for RNATargetAmount but not more than remiaingQuantity or less than RNAMinimalAmount
			rna.usedQuantity = Math.max(Math.min(rna.remainingQuantity, protocol.RNATargetAmount), protocol.RNAMinimalAmount); // µg
			rna.volume = Math.ceil(10000*rna.usedQuantity/rna.NDConc) / 10;
			rna.water = protocol.RNATargetVolume-rna.volume; // µl
			
			var remarks = [];
			if (rna.volume < MINIMAL_RNA_VOLUME)
			{
				// Large mix since we do not want to take less than 1µl
				rna.volume = MINIMAL_RNA_VOLUME;
				rna.usedQuantity = rna.volume * rna.NDConc / 1000; // µg
				var totalVolume = rna.usedQuantity / normalizedConc; // µl
				rna.water = totalVolume - rna.volume;
				remarks[remarks.length] = 'Large mix';
			}
			else if (rna.volume > protocol.RNATargetVolume)
			{
				// The RNA has too low concentration so we can't reach the target amount
				// with the target volume. The protocol may allow us to use a lesser amount
				// so we try this
				rna.volume = protocol.RNATargetVolume;
				rna.water = 0;
				rna.usedQuantity = rna.volume * rna.NDConc / 1000;
			}
			
			rna.notEnoughQuantity = rna.remainingQuantity < rna.usedQuantity || rna.usedQuantity < protocol.RNAMinimalAmount;
			if (rna.notEnoughQuantity) 
			{
				remarks[remarks.length] = '<span class="flag">Not enough quantity</span>';
			}
			else if (rna.usedQuantity < protocol.RNATargetAmount)
			{
				remarks[remarks.length] = Numbers.formatNumber(rna.usedQuantity * 1000, 0) + 'ng';
			}
			
			Doc.element('rnaVolume.'+rna.id).innerHTML = rna.notEnoughQuantity ? '' : Numbers.formatNumber(rna.volume, 1);
			Doc.element('rnaWater.'+rna.id).innerHTML = rna.notEnoughQuantity ? '' : (rna.water > 0 ? Numbers.formatNumber(rna.water, 1) : '—');
			Doc.element('remarks.'+rna.id).innerHTML = remarks.join(', ');
		}
	}
	
	normalize.insertEmpty = function(qiacubePosition)
	{
		var html = '<tr class="highlight empty">';
		html += '<td class="extract"></td>';
		html += '<td class="qiacubepos">'+qiacubePosition+'</td>';
		html += '<td class="box"></td>';
		html += '<td class="ndconc"></td>';
		html += '<td class="remain"></td>';
		html += '<td class="volume"></td>';
		html += '<td class="water"></td>';
		html += '<td class="remarks">Not selected for pre-normalization</td>';
		html += '</tr>'
		return html;
	}
	
	normalize.printVersion = function()
	{
		Wizard.hideGoNextConfirmation();
		Doc.removeClass('gonext', 'disabled');
		var frm = document.forms['reggie'];
		var reportName = 'RNA Normalization Protocol';
		var printNote = '<b>Note!</b> For better printing set page orientation to <i>portrait</i>.';
		Reggie.openPrintWindow('full-protocol', reportName, 'portrait', printNote, '../', 'normalize_extract.css');
	}

	normalize.validateStep2 = function(event)
	{}

	normalize.initializeStep3 = function()
	{
		var frm = document.forms['reggie'];
		Wizard.setCurrentStep(3);
		Doc.show('goregister');
		Doc.show('gocancel');
		Events.sendChangeEvent('normalizationDate');
		frm.normalizationDate.focus();
	}
	
	normalize.submit = function()
	{
		var frm = document.forms['reggie'];
		
		var submitInfo = {};
		submitInfo.normalizationProtocol = parseInt(frm.normalizationProtocol.value, 10);
		submitInfo.normalizationDate = frm.normalizationDate.value;
		submitInfo.normalizationOperator = frm.normalizationOperator.value;
		
		var rnaAliquots = [];
		for (var i=0; i < selectedRna.length; i++)
		{
			var rna = selectedRna[i];
			if (!rna.notEnoughQuantity)
			{
				var normalizedRna = {};
				normalizedRna.parentId = rna.id;
				normalizedRna.name = rna.nextAliquotName;

				normalizedRna.usedQuantity = rna.usedQuantity; // µg
				normalizedRna.NDConc = 1000 * rna.usedQuantity / (rna.volume+rna.water); // ng/µl
				
				rnaAliquots[rnaAliquots.length] = normalizedRna;
			}
			else
			{
				var flaggedRna = {};
				flaggedRna.id = rna.id;
				flaggedRna.flag = 'NotEnoughRemainingQuantity';
				rnaAliquots[rnaAliquots.length] = flaggedRna;
			}
		}
		
		submitInfo.rnaAliquots = rnaAliquots;

		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=RegisterNormalizedRNA';
		Wizard.showLoadingAnimation('Performing registration...');
		Wizard.asyncJsonRequest(url, normalize.submissionResults, 'POST', JSON.stringify(submitInfo));
	}

	normalize.submissionResults = function(response)
	{
		Wizard.showFinalMessage(response.messages);
		Doc.show('gorestart');
	}

	return normalize;
}();

Doc.onLoad(RnaNormalize.initPage);

